;;; -*- Mode: Lisp; -*-
;;; MD5 APR - Apache Portable Runtime version of MD5
;;; Copyright (c) 2016 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The MD5 APR algorithm is used for basic authorization by web servers Apache,
;;; Nginx, and others as well as by the htpasswd password encryption utility.
;;;
;;; Source: The Apache source repository.
;;; http://svn.apache.org/viewvc/apr/apr/trunk/crypto/apr_md5.c
;;;
;;; The algorithm is similar to the Unix version of MD5 in applying MD5 1000 times
;;; but combined with salt.

(in-package :md5)

(export 'md5-apr)

(defun shuffle-seq (seq)
 ;;; Copy of seq, which is its random perturbation.
  ;; We are marking used elements of seq with 1 in the mask array
  (let* ((len (length seq))
         (mask (yl:make-octet-vector len :initial-element 0))
         (result (make-sequence (type-of seq) len)))
    (declare (fixnum len)
             (type (simple-array (unsigned-byte 8) (*)) mask)
             (dynamic-extent mask))
    (dotimes (i len)
      (declare (fixnum i))
      (loop for j of-type fixnum = (random len)
            when (= (aref mask j) 0)					; free cell
            do (setf (elt result j) (elt seq i)				; - use it
                     (aref mask j) 1)
               (loop-finish)))
    result))

(defun md5-apr (pwd &optional (salt (subseq (shuffle-seq cl-base64::*encode-table*)
                                            0 8)))
 ;;; Args: pwd  String consisting of arbitrary ASCII characters.
  ;;       salt Pseudo-random string of 8 ASCII characters.
  #-debug (declare (optimize (speed 3) (safety 0) #+lispworks (hcl:fixnum-safety 0)))
  (flet ((%octets (string)
           (yl:string-to-octets string :external-format :ascii :null-terminate nil))
         (%concat (&rest args)
           (apply #'concatenate 'yl:simple-octet-vector args))
         (%strtr (string from to)
           ;; Translate characters in string by replacing any char from from
           ;; to the char from to with the same index.
           ;; Args: from, to  Strings of the same length
           (loop with result = (copy-seq string)
                 for i of-type fixnum upfrom 0 below (length string)
                 for pos = (yl:char-position (char string i) from)
                 when pos
                   do (setf (char result i) (char to pos))
                 finally return result)) )
    (let* ((osalt (%octets salt))
           (opwd (%octets pwd))
           (len (length opwd))
           (hash (md5:md5sum-sequence (%concat opwd osalt opwd)))
           (acc (%concat opwd (%octets #2="$apr1$") osalt)))	; accumulator octet vector
      (loop for i downfrom len by 16
            while (< 0 i)
               do (setq acc (%concat acc (subseq hash 0 (min 16 i)))))
      (loop with pwd0 = (subseq opwd 0 1)
            for i = len then (ash i -1)
            while (< 0 i)				; shift right (divide by 2)
               do (setq acc (%concat acc (if (logtest 1 i) #.(vector 0) pwd0))))
      
      (setq hash (md5:md5sum-sequence acc))
      (dotimes (i 1000)
        (declare (fixnum i))
        (let ((a (if (logtest 1 i) opwd hash)))
          (when (/= (mod i 3) 0)
            (setq a (%concat a osalt)))
          (when (/= (mod i 7) 0)
            (setq a (%concat a opwd)))
          (setq a (%concat a (if (logtest 1 i) hash opwd)))
          (setq hash (md5:md5sum-sequence a))))

      (setq acc (yl:make-octet-vector 0))
      (dotimes (i 5)
        (declare (fixnum i))
        (let ((k (+ i 6))
              (j (if (= i 4) 5 (+ i 12))))
          (declare (fixnum k j))
          (setq acc (%concat (yl:make-octet-vector 3
                                :initial-contents (list (aref hash i)
                                                        (aref hash k)
                                                        (aref hash j)))
                             acc))))
      (concatenate 'string
       #2# salt "$"
       (%strtr (reverse (subseq (cl-base64:usb8-array-to-base64-string
                                 (%concat (list 0 0 (aref hash 11)) acc))
                                2))
               cl-base64::*encode-table*
               "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"))
)) )
